﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_NQUEUE_H_
#define INCLUDE_NN_NLIB_NQUEUE_H_

#include <memory>
#include "nn/nlib/Swap.h"
#include "nn/nlib/Nlist.h"
#include "nn/nlib/TypeTraits.h"

NLIB_NAMESPACE_BEGIN

extern void* nqueue_enabler;

template <class T, class AL = std::allocator<char> >
class Nqueue {
 public:
    Nqueue() NLIB_NOEXCEPT : beg1_(0), end1_(0), beg2_(0), end2_(0) {}
    ~Nqueue() NLIB_NOEXCEPT {}
    NLIB_MOVE_MEMBER_HELPER_5(Nqueue, beg1_, end1_, beg2_, end2_, list_)
    size_t size() const NLIB_NOEXCEPT { return (end1_ - beg1_) + (end2_ - beg2_); }
    size_t capacity() const NLIB_NOEXCEPT { return list_.capacity(); }
    bool empty() const NLIB_NOEXCEPT { return end1_ - beg1_ == 0 && end2_ - beg2_ == 0; }
    T* push_back() NLIB_NOEXCEPT;
#ifdef NLIB_CXX11_RVALUE_REFERENCES
    T* push_back(T&& rhs) NLIB_NOEXCEPT;
#endif
    bool pop_front() NLIB_NOEXCEPT {
        T tmp;
        return this->pop_front_swap(&tmp);
    }
    bool push_back_swap(T* v) NLIB_NOEXCEPT {
        T* tmp = this->push_back();
        if (!tmp || !v) return false;
        using std::swap;
        swap(*tmp, *v);
        return true;
    }
    bool pop_front_swap(T* v) NLIB_NOEXCEPT;
    void swap(Nqueue& rhs) NLIB_NOEXCEPT {  // NOLINT
        using std::swap;
        swap(beg1_, rhs.beg1_);
        swap(end1_, rhs.end1_);
        swap(beg2_, rhs.beg2_);
        swap(end2_, rhs.end2_);
        swap(list_, rhs.list_);
    }

 private:
    uint32_t beg1_;
    uint32_t end1_;
    uint32_t beg2_;
    uint32_t end2_;
    Nlist<T, AL> list_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Nqueue);
};

template <class T, class AL>
T* Nqueue<T, AL>::push_back() NLIB_NOEXCEPT {
    if (end1_ != beg2_ && beg2_ != end2_) {
        using std::swap;
        swap(list_[end1_++], list_[beg2_++]);
        if (beg2_ != end2_) {
            swap(list_[end1_++], list_[beg2_++]);
        }
        if (beg2_ == end2_) {
            beg2_ = beg1_;
            end2_ = end1_;
            beg1_ = end1_ = 0;
        }
    }
    if (list_.size() == end2_) {
        T* tmp = list_.push_back();
        if (tmp) ++end2_;
        return tmp;
    } else {
        return &list_[end2_++];
    }
}

#ifdef NLIB_CXX11_RVALUE_REFERENCES
template <class T, class AL>
T* Nqueue<T, AL>::push_back(T&& rhs) NLIB_NOEXCEPT {
    if (end1_ != beg2_ && beg2_ != end2_) {
        using std::swap;
        swap(list_[end1_++], list_[beg2_++]);
        if (beg2_ != end2_) {
            swap(list_[end1_++], list_[beg2_++]);
        }
        if (beg2_ == end2_) {
            beg2_ = beg1_;
            end2_ = end1_;
            beg1_ = end1_ = 0;
        }
    }
    if (list_.size() == end2_) {
#ifdef NLIB_HAS_NATIVE_TYPETRAITS
        NLIB_STATIC_ASSERT(std::is_nothrow_move_constructible<T>::value);
#endif
        T* tmp = list_.push_back(std::move(rhs));
        if (tmp) ++end2_;
        return tmp;
    } else {
#ifdef NLIB_HAS_NATIVE_TYPETRAITS
        NLIB_STATIC_ASSERT(std::is_nothrow_move_assignable<T>::value);
#endif
        list_[end2_] = std::move(rhs);
        return &list_[end2_++];
    }
}
#endif

template <class T, class AL>
bool Nqueue<T, AL>::pop_front_swap(T* v) NLIB_NOEXCEPT {
    if (!v) return false;
    using std::swap;
    if (beg1_ != end1_) {
        swap(*v, list_[beg1_++]);
        if (beg1_ == end1_) {
            beg1_ = end1_ = 0;
        } else if ((end1_ - beg1_) * 8 < end1_) {
            uint32_t cnt = static_cast<uint32_t>(end1_ - beg1_);
            for (uint32_t i = 0; i < cnt; ++i) {
                swap(list_[i], list_[beg1_ + i]);
            }
            beg1_ = 0;
            end1_ = cnt;
        }
    } else if (beg2_ != end2_) {
        NLIB_ASSERT(beg1_ + end1_ == 0);
        swap(*v, list_[beg2_++]);
        if (beg2_ == end2_) beg2_ = end2_ = 0;
    } else {
        return false;
    }
    return true;
}

NLIB_NAMESPACE_END

#ifndef NLIB_STD_SWAP_WORKAROUND
NLIB_NAMESPACE_BEGIN
#else
NLIB_DEFINE_STD_SWAP_T_BEGIN1(std)  // NOLINT
#endif

NLIB_DEFINE_STD_SWAP_T1(AL, NLIB_NS::Nqueue)  // NOLINT

#ifndef NLIB_STD_SWAP_WORKAROUND
NLIB_NAMESPACE_END
#else
NLIB_DEFINE_STD_SWAP_T_END1(std)  // NOLINT
#endif

#endif  // INCLUDE_NN_NLIB_NQUEUE_H_
