﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_MSGPACK_JSONSTREAMPARSER_H_
#define INCLUDE_NN_NLIB_MSGPACK_JSONSTREAMPARSER_H_

#include <float.h>

#include "nn/nlib/UniquePtr.h"
#include "nn/nlib/msgpack/MpObject.h"

#if defined(_MSC_VER) && defined(nx_msgpack_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif
NLIB_NAMESPACE_BEGIN
class InputStream;
class TextReader;
namespace msgpack {

struct JsonStreamParserSettings {
    enum Format {
        FORMAT_ADAPTIVE = 0,
        FORMAT_JSON,
        FORMAT_MSGPACK
    };
    size_t max_array_size;
    size_t max_map_size;
    size_t token_buffer_size;
    size_t max_depth;
    Format format;
    bool strict;
    NLIB_CEXPR JsonStreamParserSettings() NLIB_NOEXCEPT :
        max_array_size(8192), max_map_size(256), token_buffer_size(2048), max_depth(7),
        format(FORMAT_ADAPTIVE), strict(false) {}
};

class NLIB_VIS_PUBLIC JsonStreamParser NLIB_FINAL {
 public:
    enum Event {
        NONE = -1,
        EVENT_NULL = 1,
        EVENT_TRUE,
        EVENT_FALSE,
        EVENT_STRING,
        EVENT_NUMBER_INT32,
        EVENT_NUMBER_UINT32,
        EVENT_NUMBER_INT64,
        EVENT_NUMBER_UINT64,
        EVENT_NUMBER_FLOAT,
        EVENT_NUMBER_DOUBLE,
        EVENT_START_ARRAY,
        EVENT_END_ARRAY,
        EVENT_START_MAP,
        EVENT_END_MAP,
        EVENT_KEY_NAME,
        EVENT_BINARY,  // msgpack only
        EVENT_EXT,  // msgpack only
        EVENT_END_DOCUMENT
    };
    enum Error {
        ERROR_OK = 0,
        ERROR_COLON,
        ERROR_COMMA,
        ERROR_QUOTE,
        ERROR_KEY_TOOLONG,
        ERROR_NUMBER_TOOLONG,
        ERROR_ARRAY_TOOBIG,
        ERROR_MAP_TOOBIG,
        ERROR_CHARACTER,
        ERROR_NUMBER_RANGE,
        ERROR_DEPTH,
        ERROR_TOKEN,
        ERROR_STREAM
    };
    struct Token {
        Event event;
        char* buf;
        union {
            size_t size;
            int64_t i64;
            uint64_t u64;
            double f64;
            int32_t i32;
            uint32_t u32;
            float f32;
        } number;
        bool token_toobig;
        int8_t ext;  // for EVENT_EXT
    };

 public:
    static errno_t Parse(JsonStreamParser* parser, UniquePtr<MpObject>& obj,  // NOLINT
                         bool peek) NLIB_NOEXCEPT NLIB_NONNULL_1;
    static errno_t Parse(JsonStreamParser* parser, UniquePtr<MpObject>& obj) NLIB_NOEXCEPT {  // NOLINT
        return Parse(parser, obj, false);
    }
    static errno_t Parse(UniquePtr<MpObject>& obj, const void* data, size_t n,  // NOLINT
                         const JsonStreamParserSettings& settings)
        NLIB_NOEXCEPT NLIB_NONNULL_2;
    static errno_t Parse(UniquePtr<MpObject>& obj, const void* data, size_t n) NLIB_NOEXCEPT {  // NOLINT
        JsonStreamParserSettings settings;
        return Parse(obj, data, n, settings);
    }
    static errno_t Parse(UniquePtr<MpObject>& obj, const char* str,  // NOLINT
                         const JsonStreamParserSettings& settings) NLIB_NOEXCEPT {
        size_t n = nlib_strlen(str);
        if (settings.format == JsonStreamParserSettings::FORMAT_JSON) {
            return Parse(obj, str, n, settings);
        } else {
            JsonStreamParserSettings settings_ = settings;
            settings_.format = JsonStreamParserSettings::FORMAT_JSON;
            return Parse(obj, str, n, settings_);
        }
    }
    static errno_t Parse(UniquePtr<MpObject>& obj, const char* str) NLIB_NOEXCEPT {  // NOLINT
        JsonStreamParserSettings settings;
        return Parse(obj, str, settings);
    }

    static errno_t ToInt32(const Token& token, int32_t* number) NLIB_NOEXCEPT NLIB_NONNULL;
    static errno_t ToUint32(const Token& token, uint32_t* number) NLIB_NOEXCEPT NLIB_NONNULL;
    static errno_t ToInt64(const Token& token, int64_t* number) NLIB_NOEXCEPT NLIB_NONNULL;
    static errno_t ToUint64(const Token& token, uint64_t* number) NLIB_NOEXCEPT NLIB_NONNULL;
    static errno_t ToFloat(const Token& token, float* number) NLIB_NOEXCEPT NLIB_NONNULL;
    static errno_t ToDouble(const Token& token, double* number) NLIB_NOEXCEPT NLIB_NONNULL;

    JsonStreamParser() NLIB_NOEXCEPT;
    ~JsonStreamParser() NLIB_NOEXCEPT;
    errno_t Init(const JsonStreamParserSettings& settings) NLIB_NOEXCEPT;
    errno_t Init() NLIB_NOEXCEPT {
        NLIB_CEXPR JsonStreamParserSettings settings;
        return this->Init(settings);
    }
    errno_t Open(InputStream* stream) NLIB_NOEXCEPT NLIB_NONNULL;
    errno_t Open(TextReader* reader) NLIB_NOEXCEPT NLIB_NONNULL;
    errno_t Close() NLIB_NOEXCEPT;
    Error GetError() const NLIB_NOEXCEPT;
    bool HasNext() const NLIB_NOEXCEPT;
    Event Next() NLIB_NOEXCEPT;
    bool Skip() NLIB_NOEXCEPT;
    int GetLine() const NLIB_NOEXCEPT;
    int GetColumn() const NLIB_NOEXCEPT;
    const Token& GetToken() const NLIB_NOEXCEPT { return token_; }
    NLIB_SAFE_BOOL(JsonStreamParser, error_ == ERROR_OK);

 private:
    NLIB_VIS_HIDDEN Error ReadString() NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN Error ReadNumber() NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN Event NextJson() NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN Event NextMsgpack() NLIB_NOEXCEPT;
    NLIB_VIS_HIDDEN Event NextMsgpackVal() NLIB_NOEXCEPT;

    struct JsonStreamParserPrivate;
    JsonStreamParserPrivate* prv_;
    Error error_;
    Token token_;
};

inline errno_t JsonStreamParser::ToInt32(const Token& token, int32_t* number) NLIB_NOEXCEPT {
    NLIB_EINVAL_IFNULL(number);
    switch (token.event) {
    case EVENT_NUMBER_INT32:
        *number = token.number.i32;
        break;
    case EVENT_NUMBER_UINT32:
        *number = static_cast<int32_t>(token.number.u32);
        if (token.number.u32 > INT32_MAX) return ERANGE;
        break;
    case EVENT_NUMBER_INT64:
        *number = static_cast<int32_t>(token.number.i64);
        if (token.number.i64 > INT32_MAX ||
            token.number.i64 < INT32_MIN) return ERANGE;
        break;
    case EVENT_NUMBER_UINT64:
        *number = static_cast<int32_t>(token.number.u64);
        if (token.number.u64 > INT32_MAX) return ERANGE;
        break;
    case EVENT_NUMBER_FLOAT:
        *number = static_cast<int32_t>(token.number.f32);
        if (token.number.f32 > static_cast<float>(INT32_MAX) ||
            token.number.f32 < static_cast<float>(INT32_MIN))
            return ERANGE;
        return EDOM;
    case EVENT_NUMBER_DOUBLE:
        *number = static_cast<int32_t>(token.number.f64);
        if (token.number.f64 > static_cast<double>(INT32_MAX) ||
            token.number.f64 < static_cast<double>(INT32_MIN))
            return ERANGE;
        return EDOM;
    default:
        return EINVAL;
    }
    return 0;
}

inline errno_t JsonStreamParser::ToUint32(const Token& token, uint32_t* number) NLIB_NOEXCEPT {
    NLIB_EINVAL_IFNULL(number);
    switch (token.event) {
    case EVENT_NUMBER_INT32:
        *number = static_cast<uint32_t>(token.number.i32);
        if (token.number.i32 < 0) return ERANGE;
        break;
    case EVENT_NUMBER_UINT32:
        *number = token.number.u32;
        break;
    case EVENT_NUMBER_INT64:
        *number = static_cast<uint32_t>(token.number.i64);
        if (token.number.i64 > UINT32_MAX ||
            token.number.i64 < 0) return ERANGE;
        break;
    case EVENT_NUMBER_UINT64:
        *number = static_cast<uint32_t>(token.number.u64);
        if (token.number.u64 > UINT32_MAX) return ERANGE;
        break;
    case EVENT_NUMBER_FLOAT:
        *number = static_cast<uint32_t>(token.number.f32);
        if (token.number.f32 > static_cast<float>(UINT32_MAX) ||
            token.number.f32 < 0.f)
            return ERANGE;
        return EDOM;
    case EVENT_NUMBER_DOUBLE:
        *number = static_cast<uint32_t>(token.number.f64);
        if (token.number.f64 > static_cast<double>(UINT32_MAX) ||
            token.number.f64 < 0.0)
            return ERANGE;
        return EDOM;
    default:
        return EINVAL;
    }
    return 0;
}

inline errno_t JsonStreamParser::ToInt64(const Token& token, int64_t* number) NLIB_NOEXCEPT {
    NLIB_EINVAL_IFNULL(number);
    switch (token.event) {
    case JsonStreamParser::EVENT_NUMBER_INT32:
        *number = token.number.i32;
        break;
    case JsonStreamParser::EVENT_NUMBER_UINT32:
        *number = static_cast<int64_t>(token.number.u32);
        break;
    case JsonStreamParser::EVENT_NUMBER_INT64:
        *number = token.number.i64;
        break;
    case JsonStreamParser::EVENT_NUMBER_UINT64:
        *number = static_cast<int64_t>(token.number.u64);
        if (token.number.u64 > INT64_MAX) return ERANGE;
        break;
    case EVENT_NUMBER_FLOAT:
        *number = static_cast<int64_t>(token.number.f32);
        if (token.number.f32 > static_cast<float>(INT64_MAX) ||
            token.number.f32 < static_cast<float>(INT64_MIN))
            return ERANGE;
        return EDOM;
    case EVENT_NUMBER_DOUBLE:
        *number = static_cast<int64_t>(token.number.f64);
        if (token.number.f64 > static_cast<double>(INT64_MAX) ||
            token.number.f64 < static_cast<double>(INT64_MIN))
            return ERANGE;
        return EDOM;
    default:
        return EINVAL;
    }
    return 0;
}

inline errno_t JsonStreamParser::ToUint64(const Token& token, uint64_t* number) NLIB_NOEXCEPT {
    NLIB_EINVAL_IFNULL(number);
    switch (token.event) {
    case EVENT_NUMBER_INT32:
        *number = static_cast<uint64_t>(token.number.i32);
        if (token.number.i32 < 0) return ERANGE;
        break;
    case EVENT_NUMBER_UINT32:
        *number = token.number.u32;
        break;
    case EVENT_NUMBER_INT64:
        *number = static_cast<uint64_t>(token.number.i64);
        if (token.number.i64 < 0) return ERANGE;
        break;
    case EVENT_NUMBER_UINT64:
        *number = token.number.u64;
        break;
    case EVENT_NUMBER_FLOAT:
        *number = static_cast<uint64_t>(token.number.f32);
        if (token.number.f32 > static_cast<float>(UINT64_MAX) ||
            token.number.f32 < 0.f)
            return ERANGE;
        return EDOM;
    case EVENT_NUMBER_DOUBLE:
        *number = static_cast<uint64_t>(token.number.f64);
        if (token.number.f64 > static_cast<double>(UINT64_MAX) ||
            token.number.f64 < 0.0)
            return ERANGE;
        return EDOM;
    default:
        return EINVAL;
    }
    return 0;
}

inline errno_t JsonStreamParser::ToFloat(const Token& token, float* number) NLIB_NOEXCEPT {
    NLIB_EINVAL_IFNULL(number);
    switch (token.event) {
    case EVENT_NUMBER_INT32:
        *number = static_cast<float>(token.number.i32);
        if (token.number.i32 > 16777215 || token.number.i32 < -16777215)
            return EDOM;
        break;
    case EVENT_NUMBER_UINT32:
        *number = static_cast<float>(token.number.u32);
        if (token.number.u32 > 16777215)
            return EDOM;
        break;
    case EVENT_NUMBER_INT64:
        *number = static_cast<float>(token.number.i64);
        if (token.number.i64 > 16777215 || token.number.i64 < -16777215)
            return EDOM;
        break;
    case EVENT_NUMBER_UINT64:
        *number = static_cast<float>(token.number.u64);
        if (token.number.u64 > 16777215)
            return EDOM;
        break;
    case EVENT_NUMBER_FLOAT:
        *number = token.number.f32;
        break;
    case EVENT_NUMBER_DOUBLE:
        *number = static_cast<float>(token.number.f32);
        if (token.number.f64 < FLT_MIN ||
            token.number.f64 > FLT_MAX)
            return ERANGE;
        else
            return EDOM;
    default:
        return EINVAL;
    }
    return 0;
}

inline errno_t JsonStreamParser::ToDouble(const Token& token, double* number) NLIB_NOEXCEPT {
    NLIB_EINVAL_IFNULL(number);
    switch (token.event) {
    case EVENT_NUMBER_INT32:
        *number = static_cast<double>(token.number.i32);
        break;
    case EVENT_NUMBER_UINT32:
        *number = static_cast<double>(token.number.u32);
        break;
    case EVENT_NUMBER_INT64:
        *number = static_cast<double>(token.number.i64);
        if (token.number.i64 > 9007199254740991LL ||
            token.number.i64 < -9007199254740991LL)
            return EDOM;
        break;
    case EVENT_NUMBER_UINT64:
        *number = static_cast<double>(token.number.u64);
        if (token.number.u64 > 9007199254740991ULL)
            return EDOM;
        break;
    case EVENT_NUMBER_FLOAT:
        *number = token.number.f32;
        break;
    case EVENT_NUMBER_DOUBLE:
        *number = token.number.f64;
        break;
    default:
        return EINVAL;
    }
    return 0;
}

}  // namespace msgpack
NLIB_NAMESPACE_END

#if defined(_MSC_VER) && defined(nx_msgpack_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_MSGPACK_JSONSTREAMPARSER_H_
