﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_MSGPACK_MPOBJECT_H_
#define INCLUDE_NN_NLIB_MSGPACK_MPOBJECT_H_

#include <map>
#include <utility>
#include <vector>

#include "nn/nlib/Config.h"
#include "nn/nlib/Swap.h"
#include "nn/nlib/Cstring.h"
#include "nn/nlib/Nlist.h"

#ifdef NLIB_CXX11_STDLIB_ARRAY
#include <array>  // NOLINT
#endif

#ifdef NLIB_CXX11_STDLIB_UNORDERED
#include <unordered_map>  // NOLINT
#endif

#ifdef NLIB_CXX11_STDLIB_TUPLE
#include <tuple>  // NOLINT
#endif

#if defined(_MSC_VER) && defined(nx_msgpack_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINEXPORT
#endif

NLIB_NAMESPACE_BEGIN
namespace msgpack {

class MpObject;

template <class T>
errno_t Box(MpObject* obj, const T& v);

template <class T>
errno_t Unbox(const MpObject* obj, T* v);

struct nil {};

template <class T>
inline bool operator==(const nil& lhs, const T& rhs) NLIB_NOEXCEPT {
    NLIB_UNUSED(lhs);
    NLIB_UNUSED(rhs);
    return false;
}
template <class T>
inline bool operator==(const T& lhs, const nil& rhs) NLIB_NOEXCEPT {
    NLIB_UNUSED(lhs);
    NLIB_UNUSED(rhs);
    return false;
}
inline bool operator==(const nil& lhs, const nil& rhs) NLIB_NOEXCEPT {
    NLIB_UNUSED(lhs);
    NLIB_UNUSED(rhs);
    return true;
}
template <class T>
inline bool operator!=(const nil& rhs, const T& lhs) NLIB_NOEXCEPT {
    NLIB_UNUSED(lhs);
    NLIB_UNUSED(rhs);
    return true;
}
template <class T>
inline bool operator!=(const T& rhs, const nil& lhs) NLIB_NOEXCEPT {
    NLIB_UNUSED(lhs);
    NLIB_UNUSED(rhs);
    return true;
}
inline bool operator!=(const nil& rhs, const nil& lhs) NLIB_NOEXCEPT {
    NLIB_UNUSED(lhs);
    NLIB_UNUSED(rhs);
    return false;
}

struct MpObjectKv;

class NLIB_VIS_PUBLIC MpObject NLIB_FINAL {
 private:
    union UnionType {
        uint64_t u64;
        int64_t i64;
        float f32;
        double f64;
        Nlist<MpObject>* ary;
        Nlist<MpObjectKv>* mp;
        char* str;
        void* binary;
    };

 public:
    MpObject* GetArrayItem(size_t n) NLIB_NOEXCEPT {
        MpObject* obj;
        return NLIB_LIKELY(0 == this->GetArrayItem(n, &obj)) ? obj : NULL;
    }
    const MpObject* GetArrayItem(size_t n) const NLIB_NOEXCEPT {
        const MpObject* obj;
        return NLIB_LIKELY(0 == this->GetArrayItem(n, &obj)) ? obj : NULL;
    }
    MpObject& operator[](size_t n) {
        NLIB_ASSERT(this->IsArray());
        NLIB_ASSERT(n < size_);
        return (*via_.ary)[n];
    }
    MpObject* GetMapItem(const char* str) NLIB_NOEXCEPT;
    const MpObject* GetMapItem(const char* str) const NLIB_NOEXCEPT;
    template<size_t N>
    MpObject* GetMapItem(const char (&str)[N]) NLIB_NOEXCEPT {
        return GetMapItem(&str[0]);
    }
    template<size_t N>
    const MpObject* GetMapItem(const char (&str)[N]) const NLIB_NOEXCEPT {
        return GetMapItem(&str[0]);
    }
    template<class STDSTRING>
    const MpObject* GetMapItem(const STDSTRING& str) const NLIB_NOEXCEPT {
        return GetMapItem(str.c_str());
    }
    template<class STDSTRING>
    MpObject* GetMapItem(const STDSTRING& str) NLIB_NOEXCEPT {
        return GetMapItem(str.c_str());
    }
    errno_t Resize(uint32_t n);
    MpObject* SwapArrayItem(size_t n, MpObject* obj) NLIB_NOEXCEPT {
        MpObject* p = this->GetArrayItem(n);
        if (NLIB_LIKELY(p)) p->swap(*obj);
        return p;
    }
    template<class STDSTRING>
    MpObject* SwapMapItem(const STDSTRING& key, MpObject* obj) NLIB_NOEXCEPT {
        MpObject* p = this->GetMapItem(key);
        if (NLIB_LIKELY(p)) p->swap(*obj);
        return p;
    }
    template<size_t N>
    MpObject* SwapMapItem(const char (&str)[N], MpObject* obj) NLIB_NOEXCEPT {
        return SwapMapItem(&str[0], obj);
    }
    MpObject* SwapMapItem(const char* str, MpObject* obj) NLIB_NOEXCEPT {
        MpObject* p = this->GetMapItem(str);
        if (NLIB_LIKELY(p)) p->swap(*obj);
        return p;
    }
    MpObject* AppendArrayItem() NLIB_NOEXCEPT;
    MpObjectKv* AppendMapItem() NLIB_NOEXCEPT;
    NLIB_OVERRIDE_NEW;

 public:
    MpObject() NLIB_NOEXCEPT : type_(NIL) {}  // only do minimum memory access
    ~MpObject() NLIB_NOEXCEPT { ReleaseIfNeeded(); }
    NLIB_MOVE_MEMBER_HELPER_1(MpObject, type_);

    // NOTE: implicit call was error in gcc or clang....
    template <class T>
    explicit MpObject(const T& x) {
        MyBox<T, typename IsIntegral<T>::type, typename IsSigned<T>::type>::Construct(this, x);
    }
    errno_t GetArrayItem(size_t n, MpObject** obj) NLIB_NOEXCEPT NLIB_NONNULL;
    errno_t GetArrayItem(size_t n, const MpObject** obj) const NLIB_NOEXCEPT NLIB_NONNULL;
    errno_t GetMapItem(size_t n, MpObjectKv** obj) NLIB_NOEXCEPT NLIB_NONNULL;
    errno_t GetMapItem(size_t n, const MpObjectKv** obj) const NLIB_NOEXCEPT NLIB_NONNULL;
    char* GetString() NLIB_NOEXCEPT {
        if (type_ == STRING) return via_.str;
        if (NLIB_LIKELY(type_ == STRING_INLINE)) return reinterpret_cast<char*>(&raw14data_);
        return NULL;
    }
    const char* GetString() const NLIB_NOEXCEPT {
        return const_cast<MpObject*>(this)->GetString();
    }
    void* GetBinary(uint32_t* n) NLIB_NOEXCEPT {
        if (type_ == BINARY) {
            *n = size_;
            return via_.binary;
        }
        if (NLIB_LIKELY(type_ == BINARY_INLINE)) {
            *n = raw14size_;
            return &raw14data_;
        }
        return NULL;
    }
    const void* GetBinary(uint32_t* n) const NLIB_NOEXCEPT {
        return const_cast<MpObject*>(this)->GetBinary(n);
    }
    void* GetExt(int8_t* tp, uint32_t* n) NLIB_NOEXCEPT {
        if (type_ == EXT) {
            *n = size_;
            *tp = static_cast<int8_t>(raw14data_);
            return via_.binary;
        } else {
            return NULL;
        }
    }
    const void* GetExt(int8_t* tp, uint32_t* n) const NLIB_NOEXCEPT {
        return const_cast<MpObject*>(this)->GetExt(tp, n);
    }

    errno_t InitArray(uint32_t n) NLIB_NOEXCEPT;
    errno_t InitMap(uint32_t n) NLIB_NOEXCEPT;
    errno_t InitString(uint32_t n) NLIB_NOEXCEPT;
    errno_t InitString(const char* str, uint32_t n) NLIB_NOEXCEPT;
    errno_t InitString(const char* str) NLIB_NOEXCEPT {
        uint32_t n = static_cast<uint32_t>(nlib_strlen(str));
        return InitString(str, n);
    }
    errno_t InitBinary(uint32_t n) NLIB_NOEXCEPT;
    errno_t InitBinary(const void* p, uint32_t n) NLIB_NOEXCEPT;
    errno_t InitExt(int8_t tp, uint32_t n) NLIB_NOEXCEPT;
    errno_t InitExt(int8_t tp, const void* p, uint32_t n) NLIB_NOEXCEPT;

 public:  // Box
    errno_t Box(const char* str) NLIB_NOEXCEPT NLIB_NONNULL;
    template <uint32_t n>
    errno_t Box(const char (&str)[n]) NLIB_NOEXCEPT;
    template <class T, uint32_t n>
    errno_t Box(const T (&vec)[n]);
#ifdef NLIB_CXX11_STDLIB_ARRAY
    template <class T, size_t n>
    errno_t Box(const std::array<T, n>& vec);
#endif
    template <class T>
    errno_t Box(const T& v) {
        return MyBox<T, typename IsIntegral<T>::type, typename IsSigned<T>::type>::Box(this, v);
    }
    // Use BoxBinary(p, n) or Box(reinterpret_cast<const char*>(p)) instead
    errno_t BoxBinary(const void* p, uint32_t n) NLIB_NOEXCEPT NLIB_NONNULL;
    errno_t BoxExt(int8_t tp, const void* p, uint32_t n) NLIB_NOEXCEPT NLIB_NONNULL;

 public:  // Unbox
    template <class T, size_t n>
    errno_t Unbox(T (&a)[n]) const {
        return this->Unbox(&a[0], n);
    }
#ifdef NLIB_CXX11_STDLIB_ARRAY
    template <class T, size_t n>
    errno_t Unbox(std::array<T, n>& a) {  // NOLINT
        return this->Unbox(reinterpret_cast<T*>(&*a.begin()), n);
    }
#endif
    template <size_t n>
    errno_t Unbox(char (&str)[n]) const NLIB_NOEXCEPT {
        return this->Unbox(&str[0], n);
    }
    template <class T>
    errno_t Unbox(T* a, size_t n) const NLIB_NONNULL;
    errno_t Unbox(char* str, size_t n) const NLIB_NOEXCEPT NLIB_NONNULL;
    errno_t UnboxBinary(void* p, size_t n) const NLIB_NOEXCEPT NLIB_NONNULL;
    errno_t UnboxExt(int8_t* tp, void* p, size_t n) const NLIB_NOEXCEPT NLIB_NONNULL;
    template <class T>
    errno_t Unbox(T v) const {
        // T cannot be T* though v is only pointer type
        // because of conflict against array types.
        return MyBox<typename RemovePtr<T>::type,
                     typename IsIntegral<typename RemovePtr<T>::type>::type,
                     typename IsSigned<typename RemovePtr<T>::type>::type>::Unbox(this, v);
    }

 public:
    // You can use operator=() only for basic types(never fails).
    // Use Box() otherwise.
    template <class T>
    MpObject& operator=(const T& x) {
        MyBox<T, typename IsIntegral<T>::type, typename IsSigned<T>::type>::Assign(this, x);
        return *this;
    }

    MpObject* Clone() NLIB_NOEXCEPT;
    void swap(MpObject& rhs) NLIB_NOEXCEPT {  // NOLINT
        using std::swap;
        swap(type_, rhs.type_);
        swap(raw14size_, rhs.raw14size_);
        swap(raw14data_, rhs.raw14data_);
        swap(_, rhs._);
        swap(size_, rhs.size_);
        swap(via_.u64, rhs.via_.u64);
    }

 public:
    enum ObjectType {
        NIL = 0,
        BOOLEAN,
        UINT64,
        INT64,
        FLOAT,
        DOUBLE,
        STRING,
        ARRAY,
        MAP,
        BINARY,
        EXT,
        RAW = STRING,
        STRING_INLINE = 0x80 | STRING,
        BINARY_INLINE = 0x80 | BINARY,
        RAW_INLINE = 0x80 | RAW,
        BOOLEAN_FALSE = 0x00 | BOOLEAN,
        BOOLEAN_TRUE = 0x80 | BOOLEAN
    };
    ObjectType GetType() const NLIB_NOEXCEPT {
        return static_cast<ObjectType>(type_ & 0x7F);
    }
    uint32_t GetSize() const NLIB_NOEXCEPT {
        if (type_ == STRING_INLINE || type_ == BINARY_INLINE)
            return raw14size_;
        else
            return size_;
    }
    bool IsNil() const NLIB_NOEXCEPT { return GetType() == NIL; }
    bool IsBoolean() const NLIB_NOEXCEPT { return GetType() == BOOLEAN; }
    bool IsInteger() const NLIB_NOEXCEPT {
        ObjectType tp = GetType();
        return tp == UINT64 || tp == INT64;
    }
    bool IsFloat() const NLIB_NOEXCEPT { return GetType() == FLOAT; }
    bool IsDouble() const NLIB_NOEXCEPT { return GetType() == DOUBLE; }
    bool IsString() const NLIB_NOEXCEPT { return GetType() == STRING; }
    bool IsBinary() const NLIB_NOEXCEPT { return GetType() == BINARY; }
    bool IsExt() const NLIB_NOEXCEPT { return GetType() == EXT; }
    bool IsArray() const NLIB_NOEXCEPT { return GetType() == ARRAY; }
    bool IsMap() const NLIB_NOEXCEPT { return GetType() == MAP; }

 private:
    errno_t Clone(MpObject* p) NLIB_NOEXCEPT;
    template <class T, class IS_INTEGRAL, class IS_SIGNED>
    struct MyBox {
        static errno_t Box(MpObject* This, const T& v) { return ::nlib_ns::msgpack::Box(This, v); }
        static errno_t Unbox(const MpObject* This, T* v) {
            if (!v) return EINVAL;
            return ::nlib_ns::msgpack::Unbox(This, v);
        }
        static void Assign(MpObject* This, const T& x);
        static void Construct(MpObject* This, const T& x);
    };
    void ReleaseIfNeeded() NLIB_NOEXCEPT {
        if (type_ >= STRING && type_ <= EXT) ReleaseIfNeeded_();
    }
    void ReleaseIfNeeded_() NLIB_NOEXCEPT;

 private:
    uint8_t type_;
    uint8_t raw14size_;
    uint8_t raw14data_;
    uint8_t _;
    uint32_t size_;
    UnionType via_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(MpObject);
    friend NLIB_VIS_PUBLIC bool operator==(const MpObject& lhs, const MpObject& rhs);
};

struct MpObjectKv {
    MpObject first;
    MpObject second;

 public:
    MpObjectKv() NLIB_NOEXCEPT {}
    ~MpObjectKv() NLIB_NOEXCEPT {}
    NLIB_MOVE_MEMBER_HELPER_2(MpObjectKv, first, second);
    void swap(MpObjectKv& rhs) NLIB_NOEXCEPT {  // NOLINT
        first.swap(rhs.first);
        second.swap(rhs.second);
    }

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(MpObjectKv);
};

template <>
struct MpObject::MyBox<nil, FalseType, FalseType> {
    static errno_t Box(MpObject* This, nil v) NLIB_NOEXCEPT {
        NLIB_UNUSED(v);
        This->ReleaseIfNeeded();
        This->type_ = NIL;
        return 0;
    }
    static void Assign(MpObject* This, nil x) NLIB_NOEXCEPT { This->Box(x); }
    static void Construct(MpObject* This, nil) NLIB_NOEXCEPT { This->type_ = NIL; }
};

template <>
struct MpObject::MyBox<bool, TrueType, FalseType> {
    static errno_t Box(MpObject* This, bool v) NLIB_NOEXCEPT {
        This->ReleaseIfNeeded();
        This->type_ = v ? static_cast<uint8_t>(BOOLEAN_TRUE) : static_cast<uint8_t>(BOOLEAN_FALSE);
        return 0;
    }
    static errno_t Unbox(const MpObject* This, bool* v) NLIB_NOEXCEPT {
        if (This->type_ == BOOLEAN_FALSE) {
            *v = false;
            return 0;
        } else if (NLIB_LIKELY(This->type_ == BOOLEAN_TRUE)) {
            *v = true;
            return 0;
        }
        *v = false;  // to suppress warnings
        return EACCES;
    }
    static void Assign(MpObject* This, bool x) NLIB_NOEXCEPT { This->Box(x); }
    static void Construct(MpObject* This, bool x) NLIB_NOEXCEPT {
        This->type_ = x ? static_cast<uint8_t>(BOOLEAN_TRUE) : static_cast<uint8_t>(BOOLEAN_FALSE);
    }
};

template <class T>
struct MpObject::MyBox<T, TrueType, TrueType> {
    static errno_t Box(MpObject* This, T v) NLIB_NOEXCEPT {
        This->ReleaseIfNeeded();
        This->type_ = INT64;
        This->via_.i64 = v;
        return 0;
    }
    static errno_t Unbox(const MpObject* This, T* v) NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(!v)) return EINVAL;
        if (This->type_ == INT64) {
            *v = static_cast<T>(This->via_.i64);
            return 0;
        } else if (NLIB_LIKELY(This->type_ == UINT64)) {
            *v = static_cast<T>(This->via_.u64);
            return 0;
        }
        *v = 0;  // to suppress warnings
        return EACCES;
    }
    static void Assign(MpObject* This, T x) NLIB_NOEXCEPT { This->Box(x); }
    static void Construct(MpObject* This, T x) NLIB_NOEXCEPT {
        This->type_ = INT64;
        This->via_.i64 = x;
    }
};

template <class T>
struct MpObject::MyBox<T, TrueType, FalseType> {
    static errno_t Box(MpObject* This, T v) NLIB_NOEXCEPT {
        This->ReleaseIfNeeded();
        This->type_ = UINT64;
        This->via_.u64 = v;
        return 0;
    }
    static errno_t Unbox(const MpObject* This, T* v) NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(!v)) return EINVAL;
        if (This->type_ == INT64) {
            *v = static_cast<T>(This->via_.i64);
            return 0;
        } else if (NLIB_LIKELY(This->type_ == UINT64)) {
            *v = static_cast<T>(This->via_.u64);
            return 0;
        }
        *v = 0;  // to suppress warnings
        return EACCES;
    }
    static void Assign(MpObject* This, T x) NLIB_NOEXCEPT { This->Box(x); }
    static void Construct(MpObject* This, T x) NLIB_NOEXCEPT {
        This->type_ = UINT64;
        This->via_.u64 = x;
    }
};

template <class TR1_WORKAROUND>
struct MpObject::MyBox<float, FalseType, TR1_WORKAROUND> {
    static errno_t Box(MpObject* This, float v) NLIB_NOEXCEPT {
        This->ReleaseIfNeeded();
        This->type_ = FLOAT;
        This->via_.f32 = v;
        return 0;
    }
    static errno_t Unbox(const MpObject* This, float* v) NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(!v)) return EINVAL;
        switch (This->type_) {
        case FLOAT:
            *v = This->via_.f32;
            break;
        case DOUBLE:
            *v = static_cast<float>(This->via_.f64);
            break;
        default:
            *v = 0;  // to suppress warnings
            return EACCES;
        }
        return 0;
    }
    static void Assign(MpObject* This, float x) NLIB_NOEXCEPT { This->Box(x); }
    static void Construct(MpObject* This, float x) NLIB_NOEXCEPT {
        This->type_ = FLOAT;
        This->via_.f32 = x;
    }
};

template <class TR1_WORKAROUND>
struct MpObject::MyBox<double, FalseType, TR1_WORKAROUND> {
    static errno_t Box(MpObject* This, double v) NLIB_NOEXCEPT {
        This->ReleaseIfNeeded();
        This->type_ = DOUBLE;
        This->via_.f64 = v;
        return 0;
    }
    static errno_t Unbox(const MpObject* This, double* v) NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(!v)) return EINVAL;
        switch (This->type_) {
        case FLOAT:
            *v = This->via_.f32;
            break;
        case DOUBLE:
            *v = This->via_.f64;
            break;
        default:
            *v = 0;  // to suppress warnings
            return EACCES;
        }
        return 0;
    }
    static void Assign(MpObject* This, double x) NLIB_NOEXCEPT { This->Box(x); }
    static void Construct(MpObject* This, double x) NLIB_NOEXCEPT {
        This->type_ = DOUBLE;
        This->via_.f64 = x;
    }
};

template <class STDSTRING>
struct MpObject::MyBox<STDSTRING, FalseType, FalseType> {
    static errno_t Box(MpObject* This, const STDSTRING& str) NLIB_NOEXCEPT {
        uint32_t n = static_cast<uint32_t>(str.size());
        ErrnoT e = This->InitString(n);
        if (NLIB_UNLIKELY(e != 0)) return e;
        nlib_memcpy(reinterpret_cast<void*>(This->GetString()),
                    n,
                    reinterpret_cast<const void*>(&str[0]),
                    n);
        return 0;
    }
    static errno_t Unbox(const MpObject* This, STDSTRING* str) NLIB_NOEXCEPT {
        if (NLIB_UNLIKELY(!str)) return EINVAL;
        if (NLIB_UNLIKELY(This->type_ != STRING) && NLIB_UNLIKELY(This->type_ != STRING_INLINE))
            return EACCES;
        const char* chr = This->GetString();
        NLIB_TRY { (*str).assign(chr, This->GetSize()); }
        NLIB_CATCH(std::bad_alloc&) { return ENOMEM; }
        return 0;
    }
};

template <class T, class ALLOC>
struct MpObject::MyBox<std::vector<T, ALLOC>, FalseType, FalseType> {
    static errno_t Box(MpObject* This, const std::vector<T, ALLOC>& vec) {
        uint32_t n = static_cast<uint32_t>(vec.size());
        MpObject tmp;
        ErrnoT e = tmp.InitArray(n);
        if (NLIB_UNLIKELY(e != 0)) return e;

        for (uint32_t i = 0; i < n; ++i) {
            e = (*tmp.via_.ary)[i].Box(vec[i]);
            if (NLIB_UNLIKELY(e != 0)) return e;
        }
        tmp.swap(*This);
        return 0;
    }
    static errno_t Unbox(const MpObject* This, std::vector<T, ALLOC>* vec) {
        if (NLIB_UNLIKELY(!vec)) return EINVAL;
        if (NLIB_UNLIKELY(This->type_ != ARRAY)) return EACCES;
        NLIB_TRY { vec->resize(This->size_); }
        NLIB_CATCH(std::bad_alloc&) { return ENOMEM; }
        for (uint32_t i = 0; i < This->size_; ++i) {
            ErrnoT e = (*This->via_.ary)[i].Unbox(&(*vec)[i]);
            if (NLIB_UNLIKELY(e != 0)) return e;  // *vec NOT RESET
        }
        return 0;
    }
};

template <class T, class ALLOC>
struct MpObject::MyBox<Nlist<T, ALLOC>, FalseType, FalseType> {
    static errno_t Box(MpObject* This, const Nlist<T, ALLOC>& vec) {
        size_t n = vec.size();
        MpObject tmp;
        ErrnoT e = tmp.InitArray(static_cast<uint32_t>(n));
        if (NLIB_UNLIKELY(e != 0)) return e;
        typename Nlist<T, ALLOC>::const_iterator it = vec.begin();
        for (size_t i = 0; i < n; ++it, ++i) {
            e = (*tmp.via_.ary)[i].Box(*it);
            if (NLIB_UNLIKELY(e != 0)) return e;
        }
        tmp.swap(*This);
        return 0;
    }
    static errno_t Unbox(const MpObject* This, Nlist<T, ALLOC>* vec) {
        if (NLIB_UNLIKELY(!vec)) return EINVAL;
        if (NLIB_UNLIKELY(This->type_ != ARRAY)) return EACCES;
        vec->clear();
        if (NLIB_UNLIKELY(!vec->reserve(This->size_))) return ENOMEM;
        for (uint32_t i = 0; i < This->size_; ++i) {
            T* item = vec->push_back();
            ErrnoT e = (*This->via_.ary)[i].Unbox(item);
            if (NLIB_UNLIKELY(e != 0)) return e;  // *vec NOT RESET
        }
        return 0;
    }
};

template <class T1, class T2>
struct MpObject::MyBox<std::pair<T1, T2>, FalseType, FalseType> {
    static errno_t Box(MpObject* This, const std::pair<T1, T2>& p) {
        MpObject tmp;
        ErrnoT e = tmp.InitArray(2);
        if (NLIB_UNLIKELY(e != 0)) return e;
        e = (*tmp.via_.ary)[0].Box(p.first);
        if (NLIB_UNLIKELY(e != 0)) return e;
        e = (*tmp.via_.ary)[1].Box(p.second);
        if (NLIB_UNLIKELY(e != 0)) return e;
        tmp.swap(*This);
        return 0;
    }
    static errno_t Unbox(const MpObject* This, std::pair<T1, T2>* p) {
        if (NLIB_UNLIKELY(!p)) return EINVAL;
        if (NLIB_UNLIKELY(This->type_ != ARRAY)) return EACCES;
        if (NLIB_UNLIKELY(This->size_ != 2)) return EACCES;
        ErrnoT e;
        e = (*This->via_.ary)[0].Unbox(&p->first);
        if (NLIB_UNLIKELY(e != 0)) return e;  // *p NOT RESET
        e = (*This->via_.ary)[1].Unbox(&p->second);
        if (NLIB_UNLIKELY(e != 0)) return e;  // *p NOT RESET
        return 0;
    }
};

#if defined(NLIB_CXX11_STDLIB_TUPLE) && defined(NLIB_CXX11_VARIADIC_TEMPLATES)
template <class ...Args>
struct MpObject::MyBox<std::tuple<Args...>, FalseType, FalseType > {
 private:
    template<size_t N, class HEAD>
    static errno_t BoxHelper(MpObject* tmp, const HEAD& head) {
        return (*tmp->via_.ary)[N].Box(head);
    }
    template<size_t N, class HEAD, class ...TAIL>
    static errno_t BoxHelper(MpObject* tmp, const HEAD& head, TAIL&... tail) {
        ErrnoT e = (*tmp->via_.ary)[N].Box(head);
        if (NLIB_UNLIKELY(e != 0)) return e;
        return BoxHelper<N + 1, TAIL...>(tmp, tail...);
    }
    template<size_t N, class HEAD>
    static errno_t UnboxHelper(const MpObject* This, HEAD& head) {  // NOLINT
        return (*This->via_.ary)[N].Unbox(&head);
    }
    template<size_t N, class HEAD, class ...TAIL>
    static errno_t UnboxHelper(const MpObject* This, HEAD& head, TAIL&... tail) {  // NOLINT
        ErrnoT e = (*This->via_.ary)[N].Unbox(&head);
        if (NLIB_UNLIKELY(e != 0)) return e;
        return UnboxHelper<N + 1, TAIL...>(This, tail...);
    }
    template<int ...> struct seq {};
    template<int N, int ...S> struct gens : gens<N - 1, N - 1, S...> {};
    template<int ...S>
    struct gens<0, S...> {
        typedef seq<S...> type;
    };
    template<int ...S>
    static errno_t Box(MpObject* tmp, seq<S...>, const std::tuple<Args...>& p) {
        return BoxHelper<0>(tmp, std::get<S>(p)...);
    }
    template<int ...S>
    static errno_t Unbox(const MpObject* This, seq<S...>, std::tuple<Args...>& p) {  // NOLINT
        return UnboxHelper<0>(This, std::get<S>(p)...);
    }

 public:
    static errno_t Box(MpObject* This, const std::tuple<Args...>& p) {
        size_t count = std::tuple_size<std::tuple<Args...> >::value;
        MpObject tmp;
        ErrnoT e = tmp.InitArray(static_cast<uint32_t>(count));
        if (NLIB_UNLIKELY(e != 0)) return e;
        typedef typename gens<sizeof...(Args)>::type MySeq;
        e = Box(&tmp, MySeq(), p);
        if (NLIB_UNLIKELY(e != 0)) return e;
        tmp.swap(*This);
        return 0;
    }
    static errno_t Unbox(const MpObject* This, std::tuple<Args...>* p) {
        if (NLIB_UNLIKELY(!p)) return EINVAL;
        if (NLIB_UNLIKELY(This->type_ != ARRAY)) return EACCES;
        if (NLIB_UNLIKELY(This->size_ != std::tuple_size<std::tuple<Args...> >::value))
            return EACCES;
        typedef typename gens<sizeof...(Args)>::type MySeq;
        return Unbox(This, MySeq(), *p);
    }
};
#endif

template <class K, class V, class Pr, class Alloc>
struct MpObject::MyBox<std::map<K, V, Pr, Alloc>, FalseType, FalseType> {
    static errno_t Box(MpObject* This, const std::map<K, V, Pr, Alloc>& m) {
        typedef typename std::map<K, V, Pr, Alloc> maptype;
        uint32_t n = static_cast<uint32_t>(m.size());
        MpObject tmp;
        ErrnoT e = tmp.InitMap(n);
        if (NLIB_UNLIKELY(e != 0)) return e;

        typename maptype::const_iterator it;
        typename maptype::const_iterator it_end = m.end();
        uint32_t i;
        for (i = 0, it = m.begin(); it != it_end; ++it, ++i) {
            MpObjectKv& kv = (*tmp.via_.mp)[i];
            e = kv.first.Box(it->first);
            if (NLIB_UNLIKELY(e != 0)) return e;
            e = kv.second.Box(it->second);
            if (NLIB_UNLIKELY(e != 0)) return e;
        }
        tmp.swap(*This);
        return 0;
    }
    static errno_t Unbox(const MpObject* This, std::map<K, V, Pr, Alloc>* m) {
        if (NLIB_UNLIKELY(!m)) return EINVAL;
        if (NLIB_UNLIKELY(This->type_ != MAP)) return EACCES;
        m->clear();
        for (uint32_t i = 0; i < This->size_; ++i) {
            K key;
            V value;
            const MpObjectKv& kv = (*This->via_.mp)[i];
            ErrnoT e;
            e = kv.first.Unbox(&key);
            if (NLIB_UNLIKELY(e != 0)) return e;  // *m NOT RESET
            e = kv.second.Unbox(&value);
            if (NLIB_UNLIKELY(e != 0)) return e;  // *m NOT RESET
            NLIB_TRY { (*m)[NLIB_MOVE(key)] = NLIB_MOVE(value); }
            NLIB_CATCH(...) { return ENOMEM; }
        }
        return 0;
    }
};

#ifdef NLIB_CXX11_STDLIB_UNORDERED
template <class K, class V, class Hash, class Pr, class Alloc>
struct MpObject::MyBox<std::unordered_map<K, V, Hash, Pr, Alloc>, FalseType, FalseType > {
    static errno_t Box(MpObject* This, const std::unordered_map<K, V, Hash, Pr, Alloc>& m) {
        typedef typename std::unordered_map<K, V, Hash, Pr, Alloc> maptype;
        uint32_t n = static_cast<uint32_t>(m.size());
        MpObject tmp;
        ErrnoT e = tmp.InitMap(n);
        if (NLIB_UNLIKELY(e != 0)) return e;

        typename maptype::const_iterator it;
        typename maptype::const_iterator it_end = m.end();
        uint32_t i;
        for (i = 0, it = m.begin(); it != it_end; ++it, ++i) {
            MpObjectKv& kv = (*tmp.via_.mp)[i];
            e = kv.first.Box(it->first);
            if (NLIB_UNLIKELY(e != 0)) return e;
            e = kv.second.Box(it->second);
            if (NLIB_UNLIKELY(e != 0)) return e;
        }
        tmp.swap(*This);
        return 0;
    }
    static errno_t Unbox(const MpObject* This, std::unordered_map<K, V, Hash, Pr, Alloc>* m) {
        if (NLIB_UNLIKELY(!m)) return EINVAL;
        if (NLIB_UNLIKELY(This->type_ != MAP)) return EACCES;
        m->clear();
        for (uint32_t i = 0; i < This->size_; ++i) {
            K key;
            V value;
            const MpObjectKv& kv = (*This->via_.mp)[i];
            ErrnoT e;
            e = kv.first.Unbox(&key);
            if (NLIB_UNLIKELY(e != 0)) return e;  // *m NOT RESET
            e = kv.second.Unbox(&value);
            if (NLIB_UNLIKELY(e != 0)) return e;  // *m NOT RESET
            NLIB_TRY { (*m)[NLIB_MOVE(key)] = NLIB_MOVE(value); }
            NLIB_CATCH(...) { return ENOMEM; }
        }
        return 0;
    }
};
#endif

NLIB_VIS_PUBLIC bool operator==(const MpObject& lhs, const MpObject& rhs);
inline bool operator==(const MpObjectKv& lhs, const MpObjectKv& rhs) NLIB_NOEXCEPT {
    return lhs.first == rhs.first && lhs.second == rhs.second;
}
inline bool operator!=(const MpObject& lhs, const MpObject& rhs) NLIB_NOEXCEPT {
    return !(lhs == rhs);
}
inline bool operator!=(const MpObjectKv& lhs, const MpObjectKv& rhs) NLIB_NOEXCEPT {
    return !(lhs == rhs);
}

template <uint32_t n>
errno_t MpObject::Box(const char (&str)[n]) NLIB_NOEXCEPT {
    uint32_t nn = static_cast<uint32_t>(nlib_strlen(str));
    ErrnoT e = this->InitString(nn);
    if (NLIB_UNLIKELY(e != 0)) return e;
    nlib_memcpy(this->GetString(), nn, reinterpret_cast<const void*>(&str[0]), nn);
    return 0;
}

template <class T, uint32_t n>
errno_t MpObject::Box(const T (&vec)[n]) {
    MpObject tmp;
    ErrnoT e = tmp.InitArray(n);
    if (NLIB_UNLIKELY(e != 0)) return e;
    for (uint32_t i = 0; i < n; ++i) {
        e = (*tmp.via_.ary)[i].Box(vec[i]);
        if (NLIB_UNLIKELY(e != 0)) return e;
    }
    this->swap(tmp);
    return 0;
}

#ifdef NLIB_CXX11_STDLIB_ARRAY
template <class T, size_t n>
errno_t MpObject::Box(const std::array<T, n>& vec) {
    MpObject tmp;
    ErrnoT e = tmp.InitArray(n);
    if (NLIB_UNLIKELY(e != 0)) return e;
    for (uint32_t i = 0; i < n; ++i) {
        e = (*tmp.via_.ary)[i].Box(vec[i]);
        if (NLIB_UNLIKELY(e != 0)) return e;
    }
    this->swap(tmp);
    return 0;
}
#endif

template <class T>
errno_t MpObject::Unbox(T* a, size_t n) const {
    NLIB_EINVAL_IFNULL(a);
    if (NLIB_UNLIKELY(n == 0)) return EINVAL;
    if (NLIB_UNLIKELY(type_ != ARRAY)) return EACCES;
    if (NLIB_UNLIKELY(n != size_)) return ERANGE;
    for (uint32_t i = 0; i < size_; ++i) {
        ErrnoT e = (*via_.ary)[i].Unbox(&a[i]);
        if (NLIB_UNLIKELY(e != 0)) return e;
    }
    return 0;
}

}  // namespace msgpack
NLIB_NAMESPACE_END

NLIB_DEFINE_STD_SWAP(::nlib_ns::msgpack::MpObject)
NLIB_DEFINE_STD_SWAP(::nlib_ns::msgpack::MpObjectKv)

#if defined(_MSC_VER) && defined(nx_msgpack_EXPORTS)
#undef NLIB_VIS_PUBLIC
#define NLIB_VIS_PUBLIC NLIB_WINIMPORT
#endif

#endif  // INCLUDE_NN_NLIB_MSGPACK_MPOBJECT_H_
