﻿
/*---------------------------------------------------------------------------*

  Project: CrossRoad
  Copyright (C)2012-2016 Nintendo.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

 *---------------------------------------------------------------------------*/

#pragma once
#ifndef INCLUDE_NN_NLIB_THREADING_THREAD_H_
#define INCLUDE_NN_NLIB_THREADING_THREAD_H_

#include "nn/nlib/Config.h"
#include "nn/nlib/Swap.h"
#include "nn/nlib/UniquePtr.h"
#include "nn/nlib/TypeTraits.h"
#include "nn/nlib/DateTime.h"

NLIB_NAMESPACE_BEGIN
namespace threading {
struct None {};

class ThreadSettings {
 public:
    ThreadSettings() NLIB_NOEXCEPT : is_initialized_(false) {}
    ~ThreadSettings() NLIB_NOEXCEPT {
        if (is_initialized_) {
            errno_t e = nlib_thread_attr_destroy(&attr_);
            NLIB_ASSERT_NOERR(e);
            NLIB_UNUSED(e);
        }
    }
    void SetDetachState(bool detached) NLIB_NOEXCEPT {
        this->Init_();
        errno_t e = nlib_thread_attr_setint(
            &attr_,
            NLIB_THREAD_ATTR_KEY_DETACHSTATE,
            detached ? 1 : 0);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
    }
    bool GetDetachState() const NLIB_NOEXCEPT {
        this->Init_();
        int detached;
        errno_t e = nlib_thread_attr_getint(
            &attr_,
            NLIB_THREAD_ATTR_KEY_DETACHSTATE,
            &detached);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
        return (detached != 0);
    }
    errno_t SetStackSize(int size) NLIB_NOEXCEPT {
        this->Init_();
        errno_t e = nlib_thread_attr_setint(
            &attr_,
            NLIB_THREAD_ATTR_KEY_STACKSIZE,
            size);
        return e;
    }
    int GetStackSize() const NLIB_NOEXCEPT {
        this->Init_();
        int size;
        errno_t e = nlib_thread_attr_getint(
            &attr_,
            NLIB_THREAD_ATTR_KEY_STACKSIZE,
            &size);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
        return size;
    }
    errno_t SetPriority(int priority) NLIB_NOEXCEPT {
        this->Init_();
        errno_t e = nlib_thread_attr_setint(
            &attr_,
            NLIB_THREAD_ATTR_KEY_PRIORITY,
            priority);
        return e;
    }
    int GetPriority() const NLIB_NOEXCEPT {
        this->Init_();
        int priority;
        errno_t e = nlib_thread_attr_getint(
            &attr_,
            NLIB_THREAD_ATTR_KEY_PRIORITY,
            &priority);
        NLIB_ASSERT_NOERR(e);
        NLIB_UNUSED(e);
        return priority;
    }
    nlib_thread_attr* GetPtr() NLIB_NOEXCEPT {
        return is_initialized_ ? &attr_ : NULL;
    }
    const nlib_thread_attr* GetPtr() const NLIB_NOEXCEPT {
        return is_initialized_ ? &attr_ : NULL;
    }

 private:
    void Init_() const NLIB_NOEXCEPT {
        if (!is_initialized_) {
            errno_t e = nlib_thread_attr_init(&attr_);
            NLIB_ASSERT_NOERR(e);
            NLIB_UNUSED(e);
            is_initialized_ = true;
        }
    }
    mutable bool is_initialized_;
    mutable nlib_thread_attr attr_;
};

template <class T1 = None, class T2 = None, class T3 = None, class T4 = None, class T5 = None>
struct ThreadArg {
    typedef ThreadArg<T1, T2, T3, T4, T5> ThisType;
    typedef UniquePtr<ThisType> ArgType;
    typedef void (*Func)(ArgType& ptr);  // NOLINT
    NLIB_CEXPR ThreadArg() : func(NULL), arg1(), arg2(), arg3(), arg4(), arg5() {}
    NLIB_CEXPR ThreadArg(Func func_, T1 arg1_, T2 arg2_, T3 arg3_, T4 arg4_, T5 arg5_)
        : func(func_), arg1(arg1_), arg2(arg2_), arg3(arg3_), arg4(arg4_), arg5(arg5_) {}
    static void Call(ArgType& ptr) { ptr->func(ptr); }  // NOLINT
 public:
    Func func;
    T1 arg1;
    T2 arg2;
    T3 arg3;
    T4 arg4;
    T5 arg5;
};

template <>
struct ThreadArg<None, None, None, None, None> {
    typedef ThreadArg<> ThisType;
    typedef void ArgType;
    typedef void (*Func)();
};

template <class T1>
struct ThreadArg<T1, None, None, None, None> {
    typedef ThreadArg<T1> ThisType;
    typedef UniquePtr<ThisType> ArgType;
    typedef void (*Func)(ArgType& ptr);  // NOLINT
    NLIB_CEXPR ThreadArg() : func(NULL), arg1() {}
    NLIB_CEXPR ThreadArg(Func func_, T1 arg1_) : func(func_), arg1(arg1_) {}
    static void Call(ArgType& ptr) { ptr->func(ptr); }  // NOLINT
 public:
    Func func;
    T1 arg1;
};

template <class T1, class T2>
struct ThreadArg<T1, T2, None, None, None> {
    typedef ThreadArg<T1, T2> ThisType;
    typedef UniquePtr<ThisType> ArgType;
    typedef void (*Func)(ArgType& ptr);  // NOLINT
    NLIB_CEXPR ThreadArg() : func(NULL), arg1(), arg2() {}
    NLIB_CEXPR ThreadArg(Func func_, T1 arg1_, T2 arg2_) : func(func_), arg1(arg1_), arg2(arg2_) {}
    static void Call(ArgType& ptr) { ptr->func(ptr); }  // NOLINT
 public:
    Func func;
    T1 arg1;
    T2 arg2;
};

template <class T1, class T2, class T3>
struct ThreadArg<T1, T2, T3, None, None> {
    typedef ThreadArg<T1, T2, T3> ThisType;
    typedef UniquePtr<ThisType> ArgType;
    typedef void (*Func)(ArgType& ptr);  // NOLINT
    NLIB_CEXPR ThreadArg() : func(NULL), arg1(), arg2(), arg3() {}
    NLIB_CEXPR ThreadArg(Func func_, T1 arg1_, T2 arg2_, T3 arg3_)
        : func(func_), arg1(arg1_), arg2(arg2_), arg3(arg3_) {}
    static void Call(ArgType& ptr) { ptr->func(ptr); }  // NOLINT
 public:
    Func func;
    T1 arg1;
    T2 arg2;
    T3 arg3;
};

template <class T1, class T2, class T3, class T4>
struct ThreadArg<T1, T2, T3, T4, None> {
    typedef ThreadArg<T1, T2, T3, T4> ThisType;
    typedef UniquePtr<ThisType> ArgType;
    typedef void (*Func)(ArgType& ptr);  // NOLINT
    NLIB_CEXPR ThreadArg() : func(NULL), arg1(), arg2(), arg3(), arg4() {}
    NLIB_CEXPR ThreadArg(Func func_, T1 arg1_, T2 arg2_, T3 arg3_, T4 arg4_)
        : func(func_), arg1(arg1_), arg2(arg2_), arg3(arg3_), arg4(arg4_) {}
    static void Call(ArgType& ptr) { ptr->func(ptr); }  // NOLINT
 public:
    Func func;
    T1 arg1;
    T2 arg2;
    T3 arg3;
    T4 arg4;
};

#ifdef _MSC_VER
#pragma warning(push)
#pragma warning(disable : 4180)
#endif

// code snippets:
//  Thread th;
//  void func(...) {}
//  th.Start(func, ...);     # to create a thread and run func() on it
//  th.Join();               # or th.Detach();
class NLIB_VIS_PUBLIC Thread NLIB_FINAL {
 public:
    typedef void (*ThreadFunc)(void* arg);
    NLIB_CEXPR Thread() NLIB_NOEXCEPT : thread_id_(NLIB_THREAD_INVALID) {}
    ~Thread() NLIB_NOEXCEPT {
        if (thread_id_ != NLIB_THREAD_INVALID) this->Join();
    }
    NLIB_MOVE_MEMBER_HELPER_1(Thread, thread_id_)
    errno_t StartRaw(ThreadFunc func, void* arg) NLIB_NOEXCEPT;
    errno_t StartRaw(ThreadFunc func, void* arg, const ThreadSettings& settings) NLIB_NOEXCEPT;
    errno_t Join() NLIB_NOEXCEPT {
        if (!thread_id_) return ESRCH;
        errno_t e = nlib_thread_join(thread_id_);
        if (e == 0) {
            thread_id_ = NLIB_THREAD_INVALID;
            return 0;
        } else {
            return e;
        }
    }
    errno_t Detach() NLIB_NOEXCEPT {
        errno_t e = nlib_thread_detach(thread_id_);
        if (e == 0) thread_id_ = NLIB_THREAD_INVALID;
        return e;
    }
    bool IsJoinable() const NLIB_NOEXCEPT { return thread_id_ != NLIB_THREAD_INVALID; }
    errno_t GetPriority(int32_t* priority) NLIB_NOEXCEPT {
        return nlib_thread_getpriority(this->GetNativeHandle(), reinterpret_cast<int*>(priority));
    }
    errno_t ChangePriority(int32_t priority) NLIB_NOEXCEPT {
        return nlib_thread_setpriority(this->GetNativeHandle(), priority);
    }
    errno_t SetAffinity(uint32_t affinity) NLIB_NOEXCEPT {
        return nlib_thread_setaffinity(thread_id_, affinity);
    }
    nlib_thread GetNativeHandle() const NLIB_NOEXCEPT { return thread_id_; }
    void swap(Thread& rhs) NLIB_NOEXCEPT {  // NOLINT
        using std::swap;
        swap(thread_id_, rhs.thread_id_);
    }
    bool operator==(const Thread& rhs) NLIB_NOEXCEPT {
        return nlib_thread_equal(thread_id_, rhs.thread_id_) ? true : false;
    }
    bool operator!=(const Thread& rhs) NLIB_NOEXCEPT {
        return nlib_thread_equal(thread_id_, rhs.thread_id_) ? false : true;
    }

    static errno_t YieldThread() NLIB_NOEXCEPT { return nlib_yield(); }
    static errno_t Sleep(const TimeSpan& span) NLIB_NOEXCEPT {
        return nlib_sleep(span.ToTimeValue().tick);
    }

 public:
    errno_t StartRaw(const ThreadSettings& settings, ThreadArg<>::Func func) NLIB_NOEXCEPT {
        errno_t e = this->StartRaw(Thread::Exec, reinterpret_cast<void*>(func), settings);
        return e;
    }
    errno_t StartRaw(ThreadArg<>::Func func) NLIB_NOEXCEPT {
        errno_t e = this->StartRaw(Thread::Exec, reinterpret_cast<void*>(func));
        return e;
    }
    template <class ThArg>
    errno_t StartRaw(const ThreadSettings& settings,
                     UniquePtr<ThArg>& ptr) NLIB_NOEXCEPT {  // NOLINT
        if (!ptr) return EINVAL;
        errno_t e = this->StartRaw((ThreadFunc)Thread::Exec<ThArg>,  // NOLINT
                                   reinterpret_cast<void*>(ptr.get()), settings);
        if (e == 0) ptr.release();
        return e;
    }
    template <class ThArg>
    errno_t StartRaw(UniquePtr<ThArg>& ptr) NLIB_NOEXCEPT {  // NOLINT
        if (!ptr) return EINVAL;
        errno_t e = this->StartRaw((ThreadFunc)Thread::Exec<ThArg>,  // NOLINT
                                   reinterpret_cast<void*>(ptr.get()));
        if (e == 0) ptr.release();
        return e;
    }

 private:
    static void Exec(void* p) {
        ThreadArg<>::Func f = reinterpret_cast<ThreadArg<>::Func>(p);
        f();
    }
    template <class T>
    static void Exec(void* p) {
        UniquePtr<T> args(reinterpret_cast<T*>(p));
        T::Call(args);
    }

#define NLIB_THFUNC_USESWAP_NO(tp)                                                             \
    typename EnableIf<                                                                         \
        !IsSame<None, typename RemoveCv<B>::type>::value &&                                    \
            (IsSame<FalseType, typename RemoveCv<B>::type>::value || !IsSwappable<tp>::value), \
        const tp&>::type
#define NLIB_THFUNC_USESWAP_YES(tp)                                              \
    typename EnableIf<!IsSame<None, typename RemoveCv<B>::type>::value &&        \
                          IsSame<TrueType, typename RemoveCv<B>::type>::value && \
                          IsSwappable<tp>::value,                                \
                      const tp&>::type

    template <class T>
    struct ArgType : public Conditional<
                         IsArithmetic<T>::value || IsPointer<T>::value || IsMemberPointer<T>::value,
                         FalseType, TrueType> {};

    template <class FUNC>
    struct Args0 {
        typedef typename RemoveRef<FUNC>::type FUNC_;
        template <class B>
        Args0(NLIB_THFUNC_USESWAP_NO(FUNC_) func_, B)
            : func(func_) {}
        template <class B>
        Args0(NLIB_THFUNC_USESWAP_YES(FUNC_) func_, B)
            : func() {
            using std::swap;
            swap(const_cast<FUNC&>(func_), func);
        }
        template <class B>
        Args0(typename EnableIf<IsSame<typename RemoveCv<B>::type, None>::value, const FUNC&>::type
                  func_,
              B)
            : func(func_) {}
        FUNC_ func;

     private:
        NLIB_DISALLOW_COPY_AND_ASSIGN(Args0);
    };
    template <class FUNC, class T1>
    struct Args1 : public Args0<FUNC> {
        typedef Args0<FUNC> BaseType;
        typedef typename RemoveRef<T1>::type T1_;
        template <class B>
        Args1(const FUNC& func_, NLIB_THFUNC_USESWAP_NO(T1) arg1_, B)
            : BaseType(func_, typename IsSwappable<FUNC>::type()), arg1(arg1_) {}
        template <class B>
        Args1(const FUNC& func_, NLIB_THFUNC_USESWAP_YES(T1) arg1_, B)
            : BaseType(func_, typename IsSwappable<FUNC>::type()), arg1() {
            using std::swap;
            swap(const_cast<T1&>(arg1_), arg1);
        }
        template <class B>
        Args1(typename EnableIf<IsSame<typename RemoveCv<B>::type, None>::value, const FUNC&>::type
                  func_,
              const T1& arg1_, B b)
            : BaseType(func_, b), arg1(arg1_) {}
        T1_ arg1;
    };
    template <class FUNC, class T1, class T2>
    struct Args2 : public Args1<FUNC, T1> {
        typedef Args1<FUNC, T1> BaseType;
        typedef typename RemoveRef<T2>::type T2_;
        template <class B>
        Args2(const FUNC& func_, const T1& arg1_, NLIB_THFUNC_USESWAP_NO(T2) arg2_, B)
            : BaseType(func_, arg1_, typename ArgType<T1>::type()), arg2(arg2_) {}
        template <class B>
        Args2(const FUNC& func_, const T1& arg1_, NLIB_THFUNC_USESWAP_YES(T2) arg2_, B)
            : BaseType(func_, arg1_, typename ArgType<T1>::type()), arg2() {
            using std::swap;
            swap(const_cast<T2&>(arg2_), arg2);
        }
        template <class B>
        Args2(typename EnableIf<IsSame<typename RemoveCv<B>::type, None>::value, const FUNC&>::type
                  func_,
              const T1& arg1_, const T2& arg2_, B b)
            : BaseType(func_, arg1_, b), arg2(arg2_) {}
        T2_ arg2;
    };
    template <class FUNC, class T1, class T2, class T3>
    struct Args3 : public Args2<FUNC, T1, T2> {
        typedef Args2<FUNC, T1, T2> BaseType;
        typedef typename RemoveRef<T3>::type T3_;
        template <class B>
        Args3(const FUNC& func_, const T1& arg1_, const T2& arg2_, NLIB_THFUNC_USESWAP_NO(T3) arg3_,
              B)
            : BaseType(func_, arg1_, arg2_, typename ArgType<T2>::type()), arg3(arg3_) {}
        template <class B>
        Args3(const FUNC& func, const T1& arg1_, const T2& arg2_, NLIB_THFUNC_USESWAP_YES(T3) arg3_,
              B)
            : BaseType(func, arg1_, arg2_, typename ArgType<T2>::type()), arg3() {
            using std::swap;
            swap(const_cast<T3&>(arg3_), arg3);
        }
        template <class B>
        Args3(typename EnableIf<IsSame<typename RemoveCv<B>::type, None>::value, const FUNC&>::type
                  func_,
              const T1& arg1_, const T2& arg2_, const T3& arg3_, B b)
            : BaseType(func_, arg1_, arg2_, b), arg3(arg3_) {}
        T3_ arg3;
    };
    template <class FUNC, class T1, class T2, class T3, class T4>
    struct Args4 : public Args3<FUNC, T1, T2, T3> {
        typedef Args3<FUNC, T1, T2, T3> BaseType;
        typedef typename RemoveRef<T4>::type T4_;
        template <class B>
        Args4(const FUNC& func_, const T1& arg1_, const T2& arg2_, const T3& arg3_,
              NLIB_THFUNC_USESWAP_NO(T4) arg4_, B)
            : BaseType(func_, arg1_, arg2_, arg3_, typename ArgType<T3>::type()), arg4(arg4_) {}
        template <class B>
        Args4(const FUNC& func_, const T1& arg1_, const T2& arg2_, const T3& arg3_,
              NLIB_THFUNC_USESWAP_YES(T4) arg4_, B)
            : BaseType(func_, arg1_, arg2_, arg3_, typename ArgType<T3>::type()), arg4() {
            using std::swap;
            swap(const_cast<T4&>(arg4_), arg4);
        }
        template <class B>
        Args4(typename EnableIf<IsSame<typename RemoveCv<B>::type, None>::value, const FUNC&>::type
                  func_,
              const T1& arg1_, const T2& arg2_, const T3& arg3_, const T4& arg4_, B b)
            : BaseType(func_, arg1_, arg2_, arg3_, b), arg4(arg4_) {}
        T4_ arg4;
    };
    template <class FUNC, class T1, class T2, class T3, class T4, class T5>
    struct Args5 : public Args4<FUNC, T1, T2, T3, T4> {
        typedef Args4<FUNC, T1, T2, T3, T4> BaseType;
        typedef typename RemoveRef<T5>::type T5_;
        template <class B>
        Args5(const FUNC& func_, const T1& arg1_, const T2& arg2_, const T3& arg3_, const T4& arg4_,
              NLIB_THFUNC_USESWAP_NO(T5) arg5_, B)
            : BaseType(func_, arg1_, arg2_, arg3_, arg4_, typename ArgType<T4>::type()),
              arg5(arg5_) {}
        template <class B>
        Args5(const FUNC& func_, const T1& arg1_, const T2& arg2_, const T3& arg3_, const T4& arg4_,
              NLIB_THFUNC_USESWAP_YES(T4) arg5_, B)
            : BaseType(func_, arg1_, arg2_, arg3_, arg4_, typename ArgType<T4>::type()), arg5() {
            using std::swap;
            swap(const_cast<T5&>(arg5_), arg5);
        }
        template <class B>
        Args5(typename EnableIf<IsSame<typename RemoveCv<B>::type, None>::value, const FUNC&>::type
                  func_,
              const T1& arg1_, const T2& arg2_, const T3& arg3_, const T4& arg4_, const T5& arg5_,
              B b)
            : BaseType(func_, arg1_, arg2_, arg3_, arg4_, b), arg5(arg5_) {}
        T5_ arg5;
    };
#undef NLIB_THFUNC_USESWAP_NO
#undef NLIB_THFUNC_USESWAP_YES

    template <class FUNC>
    struct ThreadFuncX0 : public Args0<FUNC> {
        typedef ThreadFuncX0<FUNC> ThisType;
        typedef Args0<FUNC> BaseType;
        template <class B>
        ThreadFuncX0(const FUNC& func_, B b)
            : BaseType(func_, b) {}
        explicit ThreadFuncX0(const FUNC& func_) : BaseType(func_, None()) {}
        static void Call(UniquePtr<ThisType>& ptr) {  // NOLINT
            ptr->func();
        }
    };
    template <class FUNC, class T1>
    struct ThreadFuncX1 : public Args1<FUNC, T1> {
        typedef ThreadFuncX1<FUNC, T1> ThisType;
        typedef Args1<FUNC, T1> BaseType;
        template <class B>
        ThreadFuncX1(const FUNC& func_, const T1& arg1_, B b)
            : BaseType(func_, arg1_, b) {}
        ThreadFuncX1(const FUNC& func_, const T1& arg1_) : BaseType(func_, arg1_, None()) {}
        static void Call(UniquePtr<ThisType>& ptr) {  // NOLINT
            ptr->func(NLIB_MOVE(ptr->arg1));
        }
    };
    template <class FUNC, class T1, class T2>
    struct ThreadFuncX2 : public Args2<FUNC, T1, T2> {
        typedef ThreadFuncX2<FUNC, T1, T2> ThisType;
        typedef Args2<FUNC, T1, T2> BaseType;
        template <class B>
        ThreadFuncX2(const FUNC& func_, const T1& arg1_, const T2& arg2_, B b)
            : BaseType(func_, arg1_, arg2_, b) {}
        ThreadFuncX2(const FUNC& func_, const T1& arg1_, const T2& arg2_)
            : BaseType(func_, arg1_, arg2_, None()) {}
        static void Call(UniquePtr<ThisType>& ptr) {  // NOLINT
            ptr->func(NLIB_MOVE(ptr->arg1), NLIB_MOVE(ptr->arg2));
        }
    };
    template <class FUNC, class T1, class T2, class T3>
    struct ThreadFuncX3 : public Args3<FUNC, T1, T2, T3> {
        typedef ThreadFuncX3<FUNC, T1, T2, T3> ThisType;
        typedef Args3<FUNC, T1, T2, T3> BaseType;
        template <class B>
        ThreadFuncX3(const FUNC& func_, const T1& arg1_, const T2& arg2_, const T3& arg3_, B b)
            : BaseType(func_, arg1_, arg2_, arg3_, b) {}
        ThreadFuncX3(const FUNC& func_, const T1& arg1_, const T2& arg2_, const T3& arg3_)
            : BaseType(func_, arg1_, arg2_, arg3_, None()) {}
        static void Call(UniquePtr<ThisType>& ptr) {  // NOLINT
            ptr->func(NLIB_MOVE(ptr->arg1), NLIB_MOVE(ptr->arg2), NLIB_MOVE(ptr->arg3));
        }
    };
    template <class FUNC, class T1, class T2, class T3, class T4>
    struct ThreadFuncX4 : public Args4<FUNC, T1, T2, T3, T4> {
        typedef ThreadFuncX4<FUNC, T1, T2, T3, T4> ThisType;
        typedef Args4<FUNC, T1, T2, T3, T4> BaseType;
        template <class B>
        ThreadFuncX4(const FUNC& func_, const T1& arg1_, const T2& arg2_, const T3& arg3_,
                     const T4& arg4_, B b)
            : BaseType(func_, arg1_, arg2_, arg3_, arg4_, b) {}
        ThreadFuncX4(const FUNC& func_, const T1& arg1_, const T2& arg2_, const T3& arg3_,
                     const T4& arg4_)
            : BaseType(func_, arg1_, arg2_, arg3_, arg4_, None()) {}
        static void Call(UniquePtr<ThisType>& ptr) {  // NOLINT
            ptr->func(NLIB_MOVE(ptr->arg1), NLIB_MOVE(ptr->arg2), NLIB_MOVE(ptr->arg3),
                      NLIB_MOVE(ptr->arg4));
        }
    };
    template <class FUNC, class T1, class T2, class T3, class T4, class T5>
    struct ThreadFuncX5 : public Args5<FUNC, T1, T2, T3, T4, T5> {
        typedef ThreadFuncX5<FUNC, T1, T2, T3, T4, T5> ThisType;
        typedef Args5<FUNC, T1, T2, T3, T4, T5> BaseType;
        template <class B>
        ThreadFuncX5(const FUNC& func_, const T1& arg1_, const T2& arg2_, const T3& arg3_,
                     const T4& arg4_, const T5& arg5_, B b)
            : BaseType(func_, arg1_, arg2_, arg3_, arg4_, arg5_, b) {}
        ThreadFuncX5(const FUNC& func_, const T1& arg1_, const T2& arg2_, const T3& arg3_,
                     const T4& arg4_, const T5& arg5_)
            : BaseType(func_, arg1_, arg2_, arg3_, arg4_, arg5_, None()) {}
        static void Call(UniquePtr<ThisType>& ptr) {  // NOLINT
            ptr->func(NLIB_MOVE(ptr->arg1), NLIB_MOVE(ptr->arg2), NLIB_MOVE(ptr->arg3),
                      NLIB_MOVE(ptr->arg4), NLIB_MOVE(ptr->arg5));
        }
    };

 public:
    template <class FUNC>
    errno_t Start(const ThreadSettings& settings, const FUNC& f, move_tag) {
        UniquePtr<ThreadFuncX0<FUNC> > ptr(
            new (std::nothrow) ThreadFuncX0<FUNC>(f, typename IsSwappable<FUNC>::type()));
        if (!ptr) return ENOMEM;
        return this->StartRaw(settings, ptr);
    }
    template <class FUNC>
    errno_t Start(const FUNC& f, move_tag) {
        UniquePtr<ThreadFuncX0<FUNC> > ptr(
            new (std::nothrow) ThreadFuncX0<FUNC>(f, typename IsSwappable<FUNC>::type()));
        if (!ptr) return ENOMEM;
        return this->StartRaw(ptr);
    }
    template <class FUNC, class T1>
    errno_t Start(const ThreadSettings& settings, const FUNC& f, const T1& arg1, move_tag) {
        UniquePtr<ThreadFuncX1<FUNC, T1> > ptr(
            new (std::nothrow) ThreadFuncX1<FUNC, T1>(f, arg1, typename ArgType<T1>::type()));
        if (!ptr) return ENOMEM;
        return this->StartRaw(settings, ptr);
    }
    template <class FUNC, class T1>
    errno_t Start(const FUNC& f, const T1& arg1, move_tag) {
        UniquePtr<ThreadFuncX1<FUNC, T1> > ptr(
            new (std::nothrow) ThreadFuncX1<FUNC, T1>(f, arg1, typename ArgType<T1>::type()));
        if (!ptr) return ENOMEM;
        return this->StartRaw(ptr);
    }
    template <class FUNC, class T1, class T2>
    errno_t Start(const ThreadSettings& settings, const FUNC& f, const T1& arg1, const T2& arg2,
                  move_tag) {
        UniquePtr<ThreadFuncX2<FUNC, T1, T2> > ptr(new (std::nothrow) ThreadFuncX2<FUNC, T1, T2>(
            f, arg1, arg2, typename ArgType<T2>::type()));
        if (!ptr) return ENOMEM;
        return this->StartRaw(settings, ptr);
    }
    template <class FUNC, class T1, class T2>
    errno_t Start(const FUNC& f, const T1& arg1, const T2& arg2, move_tag) {
        UniquePtr<ThreadFuncX2<FUNC, T1, T2> > ptr(new (std::nothrow) ThreadFuncX2<FUNC, T1, T2>(
            f, arg1, arg2, typename ArgType<T2>::type()));
        if (!ptr) return ENOMEM;
        return this->StartRaw(ptr);
    }
    template <class FUNC, class T1, class T2, class T3>
    errno_t Start(const ThreadSettings& settings, const FUNC& f, const T1& arg1, const T2& arg2,
                  const T3& arg3, move_tag) {
        UniquePtr<ThreadFuncX3<FUNC, T1, T2, T3> > ptr(
            new (std::nothrow)
            ThreadFuncX3<FUNC, T1, T2, T3>(f, arg1, arg2, arg3, typename ArgType<T3>::type()));
        if (!ptr) return ENOMEM;
        return this->StartRaw(settings, ptr);
    }
    template <class FUNC, class T1, class T2, class T3>
    errno_t Start(const FUNC& f, const T1& arg1, const T2& arg2, const T3& arg3, move_tag) {
        UniquePtr<ThreadFuncX3<FUNC, T1, T2, T3> > ptr(
            new (std::nothrow)
            ThreadFuncX3<FUNC, T1, T2, T3>(f, arg1, arg2, arg3, typename ArgType<T3>::type()));
        if (!ptr) return ENOMEM;
        return this->StartRaw(ptr);
    }
    template <class FUNC, class T1, class T2, class T3, class T4>
    errno_t Start(const ThreadSettings& settings, const FUNC& f, const T1& arg1, const T2& arg2,
                  const T3& arg3, const T4& arg4, move_tag) {
        UniquePtr<ThreadFuncX4<FUNC, T1, T2, T3, T4> > ptr(
            new (std::nothrow) ThreadFuncX4<FUNC, T1, T2, T3, T4>(f, arg1, arg2, arg3, arg4,
                                                                  typename ArgType<T4>::type()));
        if (!ptr) return ENOMEM;
        return this->StartRaw(settings, ptr);
    }
    template <class FUNC, class T1, class T2, class T3, class T4>
    errno_t Start(const FUNC& f, const T1& arg1, const T2& arg2, const T3& arg3, const T4& arg4,
                  move_tag) {
        UniquePtr<ThreadFuncX4<FUNC, T1, T2, T3, T4> > ptr(
            new (std::nothrow) ThreadFuncX4<FUNC, T1, T2, T3, T4>(f, arg1, arg2, arg3, arg4,
                                                                  typename ArgType<T4>::type()));
        if (!ptr) return ENOMEM;
        return this->StartRaw(ptr);
    }
    template <class FUNC, class T1, class T2, class T3, class T4, class T5>
    errno_t Start(const ThreadSettings& settings, const FUNC& f, const T1& arg1, const T2& arg2,
                  const T3& arg3, const T4& arg4, const T5& arg5, move_tag) {
        UniquePtr<ThreadFuncX5<FUNC, T1, T2, T3, T4, T5> > ptr(
            new (std::nothrow) ThreadFuncX5<FUNC, T1, T2, T3, T4, T5>(
                f, arg1, arg2, arg3, arg4, arg5, typename ArgType<T5>::type()));
        if (!ptr) return ENOMEM;
        return this->StartRaw(settings, ptr);
    }
    template <class FUNC, class T1, class T2, class T3, class T4, class T5>
    errno_t Start(const FUNC& f, const T1& arg1, const T2& arg2, const T3& arg3, const T4& arg4,
                  const T5& arg5, move_tag) {
        UniquePtr<ThreadFuncX5<FUNC, T1, T2, T3, T4, T5> > ptr(
            new (std::nothrow) ThreadFuncX5<FUNC, T1, T2, T3, T4, T5>(
                f, arg1, arg2, arg3, arg4, arg5, typename ArgType<T5>::type()));
        if (!ptr) return ENOMEM;
        return this->StartRaw(ptr);
    }
    template <class FUNC>
    errno_t Start(const ThreadSettings& settings, const FUNC& f) {
        UniquePtr<ThreadFuncX0<FUNC> > ptr(new (std::nothrow) ThreadFuncX0<FUNC>(f));
        if (!ptr) return ENOMEM;
        return this->StartRaw(settings, ptr);
    }
    template <class FUNC>
    errno_t Start(const FUNC& f) {
        UniquePtr<ThreadFuncX0<FUNC> > ptr(new (std::nothrow) ThreadFuncX0<FUNC>(f));
        if (!ptr) return ENOMEM;
        return this->StartRaw(ptr);
    }
    template <class FUNC, class T1>
    errno_t Start(const ThreadSettings& settings, const FUNC& f, const T1& arg1) {
        UniquePtr<ThreadFuncX1<FUNC, T1> > ptr(new (std::nothrow) ThreadFuncX1<FUNC, T1>(f, arg1));
        if (!ptr) return ENOMEM;
        return this->StartRaw(settings, ptr);
    }
    template <class FUNC, class T1>
    errno_t Start(const FUNC& f, const T1& arg1) {
        UniquePtr<ThreadFuncX1<FUNC, T1> > ptr(new (std::nothrow) ThreadFuncX1<FUNC, T1>(f, arg1));
        if (!ptr) return ENOMEM;
        return this->StartRaw(ptr);
    }
    template <class FUNC, class T1, class T2>
    errno_t Start(const ThreadSettings& settings, const FUNC& f, const T1& arg1, const T2& arg2) {
        UniquePtr<ThreadFuncX2<FUNC, T1, T2> > ptr(new (std::nothrow)
                                                   ThreadFuncX2<FUNC, T1, T2>(f, arg1, arg2));
        if (!ptr) return ENOMEM;
        return this->StartRaw(settings, ptr);
    }
    template <class FUNC, class T1, class T2>
    errno_t Start(const FUNC& f, const T1& arg1, const T2& arg2) {
        UniquePtr<ThreadFuncX2<FUNC, T1, T2> > ptr(new (std::nothrow)
                                                   ThreadFuncX2<FUNC, T1, T2>(f, arg1, arg2));
        if (!ptr) return ENOMEM;
        return this->StartRaw(ptr);
    }
    template <class FUNC, class T1, class T2, class T3>
    errno_t Start(const ThreadSettings& settings, const FUNC& f, const T1& arg1, const T2& arg2,
                  const T3& arg3) {
        UniquePtr<ThreadFuncX3<FUNC, T1, T2, T3> > ptr(
            new (std::nothrow) ThreadFuncX3<FUNC, T1, T2, T3>(f, arg1, arg2, arg3));
        if (!ptr) return ENOMEM;
        return this->StartRaw(settings, ptr);
    }
    template <class FUNC, class T1, class T2, class T3>
    errno_t Start(const FUNC& f, const T1& arg1, const T2& arg2, const T3& arg3) {
        UniquePtr<ThreadFuncX3<FUNC, T1, T2, T3> > ptr(
            new (std::nothrow) ThreadFuncX3<FUNC, T1, T2, T3>(f, arg1, arg2, arg3));
        if (!ptr) return ENOMEM;
        return this->StartRaw(ptr);
    }
    template <class FUNC, class T1, class T2, class T3, class T4>
    errno_t Start(const ThreadSettings& settings, const FUNC& f, const T1& arg1, const T2& arg2,
                  const T3& arg3, const T4& arg4) {
        UniquePtr<ThreadFuncX4<FUNC, T1, T2, T3, T4> > ptr(
            new (std::nothrow) ThreadFuncX4<FUNC, T1, T2, T3, T4>(f, arg1, arg2, arg3, arg4));
        if (!ptr) return ENOMEM;
        return this->StartRaw(settings, ptr);
    }
    template <class FUNC, class T1, class T2, class T3, class T4>
    errno_t Start(const FUNC& f, const T1& arg1, const T2& arg2, const T3& arg3, const T4& arg4) {
        UniquePtr<ThreadFuncX4<FUNC, T1, T2, T3, T4> > ptr(
            new (std::nothrow) ThreadFuncX4<FUNC, T1, T2, T3, T4>(f, arg1, arg2, arg3, arg4));
        if (!ptr) return ENOMEM;
        return this->StartRaw(ptr);
    }
    template <class FUNC, class T1, class T2, class T3, class T4, class T5>
    errno_t Start(const ThreadSettings& settings, const FUNC& f, const T1& arg1, const T2& arg2,
                  const T3& arg3, const T4& arg4, const T5& arg5) {
        UniquePtr<ThreadFuncX5<FUNC, T1, T2, T3, T4, T5> > ptr(new (
            std::nothrow) ThreadFuncX5<FUNC, T1, T2, T3, T4, T5>(f, arg1, arg2, arg3, arg4, arg5));
        if (!ptr) return ENOMEM;
        return this->StartRaw(settings, ptr);
    }
    template <class FUNC, class T1, class T2, class T3, class T4, class T5>
    errno_t Start(const FUNC& f, const T1& arg1, const T2& arg2, const T3& arg3, const T4& arg4,
                  const T5& arg5) {
        UniquePtr<ThreadFuncX5<FUNC, T1, T2, T3, T4, T5> > ptr(new (
            std::nothrow) ThreadFuncX5<FUNC, T1, T2, T3, T4, T5>(f, arg1, arg2, arg3, arg4, arg5));
        if (!ptr) return ENOMEM;
        return this->StartRaw(ptr);
    }

 private:
    nlib_thread thread_id_;
    NLIB_DISALLOW_COPY_AND_ASSIGN(Thread);
};

#ifdef _MSC_VER
#pragma warning(pop)
#endif

template <>
struct Thread::Args0<void()> {
    template <class X>
    Args0(void (*func_)(), X)
        : func(func_) {}
    Args0(void (*func_)()) : func(func_) {}
    void (*func)();

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(Args0);
};

template <class T1>
struct Thread::Args0<void(T1)> {
    template <class X>
    Args0(void (*func_)(T1), X)
        : func(func_) {}
    Args0(void (*func_)(T1)) : func(func_) {}
    void (*func)(T1);

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(Args0);
};

template <class T1, class T2>
struct Thread::Args0<void(T1, T2)> {
    template <class X>
    Args0(void (*func_)(T1, T2), X)
        : func(func_) {}
    Args0(void (*func_)(T1, T2)) : func(func_) {}
    void (*func)(T1, T2);

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(Args0);
};

template <class T1, class T2, class T3>
struct Thread::Args0<void(T1, T2, T3)> {
    template <class X>
    Args0(void (*func_)(T1, T2, T3), X)
        : func(func_) {}
    Args0(void (*func_)(T1, T2, T3)) : func(func_) {}
    void (*func)(T1, T2, T3);

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(Args0);
};

template <class T1, class T2, class T3, class T4>
struct Thread::Args0<void(T1, T2, T3, T4)> {
    template <class X>
    Args0(void (*func_)(T1, T2, T3, T4), X)
        : func(func_) {}
    Args0(void (*func_)(T1, T2, T3, T4)) : func(func_) {}
    void (*func)(T1, T2, T3, T4);

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(Args0);
};

template <class T1, class T2, class T3, class T4, class T5>
struct Thread::Args0<void(T1, T2, T3, T4, T5)> {
    template <class X>
    Args0(void (*func_)(T1, T2, T3, T4, T5), X)
        : func(func_) {}
    Args0(void (*func_)(T1, T2, T3, T4, T5)) : func(func_) {}
    void (*func)(T1, T2, T3, T4, T5);

 private:
    NLIB_DISALLOW_COPY_AND_ASSIGN(Args0);
};

inline size_t GetHardwareConcurrency() NLIB_NOEXCEPT {
    unsigned int numCpu;
    (void)nlib_thread_getconcurrency(&numCpu);
    return numCpu;
}

namespace this_thread {
inline errno_t YieldThread() NLIB_NOEXCEPT { return nlib_yield(); }
inline errno_t Sleep(const TimeSpan& span) NLIB_NOEXCEPT {
    return ::nlib_ns::threading::Thread::Sleep(span);
}
inline errno_t SleepMilliSeconds(unsigned int millisec) NLIB_NOEXCEPT {
    return Sleep(TimeSpan(0, 0, millisec));
}
inline nlib_thread_id GetId() NLIB_NOEXCEPT {
    nlib_thread_id id;
    errno_t e = nlib_thread_getid(&id);
    return e == 0 ? id : -1;
}
inline errno_t GetCpu(int* cpuid) NLIB_NOEXCEPT { return nlib_thread_getcpu(cpuid); }
inline errno_t SetName(const char* literal_string) NLIB_NOEXCEPT {
    nlib_thread th;
    errno_t e = nlib_thread_self(&th);
    if (e != 0) return e;
    return nlib_thread_setname(th, literal_string);
}
NLIB_VIS_PUBLIC errno_t GetPriority(int32_t* priority) NLIB_NOEXCEPT;
NLIB_VIS_PUBLIC errno_t ChangePriority(int32_t priority) NLIB_NOEXCEPT;

}  // namespace this_thread

}  // namespace threading
NLIB_NAMESPACE_END

NLIB_DEFINE_STD_SWAP(::nlib_ns::threading::Thread)

#endif  // INCLUDE_NN_NLIB_THREADING_THREAD_H_
